/*
 * Copyright Francesco Potortì 2014
 *
 * This software is distributed under the GPLv3 licence
 */

#include "enerduino.h"

/************************ Global variables **************************/

char ssid[] = "casapot";		// name of wifi network

bool usingSD;				// whether we have an SD card
int wifiStatus;				// status of WiFi connection to a network

const millis_t wifidogAdvance = 200000;	  // need a connection at least every this many ms
unsigned long long unixTimeAtStartup;	  // milliseconds since 19700101T0000
millis_t wifidogTimeout = wifidogAdvance; // countdown for inactive WiFi server (seconds)

// corLen divided by the squared sum of prototypes
unsigned long BlinkSensor::protoMult[BlinkSensor::protoNum];
// The prototoype blink, in protoNum versions
byte *BlinkSensor::proto;

#define debugLed false			// DEBUG: service led for blinking
#if debugLed
  const int debugLED = 2;
  const long debugLedPeriod = 1800;
  long debugLedFlashLen;
  millis_t debugLedOnTime;
#endif

unsigned taskStackLeft;			// DEBUG: stack left unused by the webServer task
byte *taskStackBase;			// DEBUG: used to compute unused stack
unsigned int taskStackSize = 200;

int lateReadMax[2];			// DEBUG: late read time
unsigned maxt;
#define timecount									\
  for (unsigned start = 0, t = 0;							\
     start ? false : (t = 0, start = micros(), t -= micros() - start, start = micros());\
     t += micros() - start, maxt = max(maxt, t))

// A FIFO to store sensor readings
struct reading {
  char type;
  unsigned long long time;
  int val;
};
Fifo<reading, big> readings;
unsigned readingLost;			// FIFO full error count

bool disaster;				// max correlation problem


/******************** Forward function declarations *******************/

void mainLoopScheduler(AnalogSensor *[], AnalogSensor *[], const byte,
		       AnalogSensor *&, AnalogSensor *&);
void webServerTask ();

void ledOnOff (int pin, unsigned on = HIGH, unsigned off = LOW);
time_t webUnixTime (Client&);
byte *startLoop (SchedulerTask, unsigned);
extern byte *__brkval;			// set after malloc, it is the top of the heap
unsigned stackUnused(byte * = __brkval, unsigned = 0);
void resetDebugStats ();
void resetWiFiShield ();
void resetArduino ();

/**************** Main entry points: loop and setup *****************/

void setup () {
  // Reduce malloc margin from the default of 32
  __malloc_margin = 16;

  // Needed for bootloaders that do not disable the watchdog timer
  if (usingWatchdogTimer)
    wdt_disable();

  // Reset the WiFi shield and wait for it to start
  resetWiFiShield();
  delay(500);

  // Debugging: start the serial interface, use leds, init unused stack measurement
  Serial.begin(115200);
  pinMode(internalLED, OUTPUT);
# if debugLed
    pinMode(debugLED, OUTPUT);
# endif

  // Initialize SD card
  //  usingSD = SD.begin(4);
  // Check for WiFi shield presence
  wifiStatus = WiFi.status();

  // Start the tasks
  taskStackBase = startLoop(webServerTask, taskStackSize);

  // Announce initialisation finished
  Serial << (usingSD ? "" : "no ") << F("SD present, ")
	 << (wifiStatus == WL_NO_SHIELD ? "no" : "") << F(" WiFi shield present, ")
	 << F("free mem: ") << ({byte a; &a - __brkval;}) << endl; delay(50);

  // Unchain the watch dog
  if (usingWatchdogTimer)
    wdt_enable(WDTO_15MS);
}

void loop ()
{
  static AnalogSensor *blinkSensor[2] = {
    new BlinkSensor(0),
    reactiveSensorUsed ? new BlinkSensor(1) : 0
  };
  // At least one other sensor must be there
  static AnalogSensor *otherSensor[] = {
    new InternalTempSensor(8)		// 8 is the MUX input used for the internal temp sensor
  };
  const byte numOther = sizeof(otherSensor)/sizeof(*otherSensor);
  static AnalogSensor *csp = otherSensor[0]; // current sensor index
  static AnalogSensor *nsp = blinkSensor[0]; // next sensor index

  enum status { reset, Reset, debug, Debug, reading, idle };
  static status st = Reset;


  // Signal that we are still alive
  if (usingWatchdogTimer)
    wdt_reset();			// pat the watch dog

  // If fatal error, do nothing but keep the web server working to signal it
  if (disaster)
    { yield(); return; }

  // Visual LED heartbeat
  if (wifiStatus != WL_CONNECTED)	// we are not connected to a WiFi network
    ledOnOff(internalLED, 50, 2000);
  else if (millis() - wifidogTimeout < 60000) // one minute left before wifidog bites
    ledOnOff(internalLED, 50, 1000);
  else					// we are connected and time is set
    ledOnOff(internalLED, 50, 500);

  // DEBUG: service led for blinking
# if debugLed
    if (debugLedFlashLen < 0)
      ledOnOff (debugLED, 20, -20-debugLedFlashLen);
    else if (debugLedFlashLen > 0
	   && millis() - debugLedOnTime > (millis_t)debugLedFlashLen)
    {
	ledOnOff (debugLED, LOW);
	debugLedFlashLen = 0;
      }
# endif

  // Wifi dog management
  if (millis() > wifidogTimeout)	// don't wait long without a connection
    resetWiFiShield();			// wifi shield stuck...


  if (Serial.available())		// DEBUG
    switch (Serial.read())
      {
      case 'm': st = reset; break;
      case 'r': st = reading; Serial << F("Reading\r\n"); break;
      case 'i': st = idle; break;
#   if debugLed
      case 'f': ledOnOff(debugLED, HIGH); debugLedOnTime = millis(); debugLedFlashLen = 20; break;
      case 'F': ledOnOff(debugLED, HIGH); debugLedOnTime = millis(); debugLedFlashLen = 60ul*60*1000; break;
#   endif
      case 'D': st = Debug; break;
      case 'd': st = debug; break;
      case 'y': yield(); break;
      case 'R': resetArduino(); break;	// restart from scratch

      }

  switch (st)				// DEBUG
    {
    case idle:
      digitalWrite(internalLED, HIGH);
      return;				// don't read, don't yield

    case Debug:
    case debug:
      Serial << F("Stack: Main ") << stackUnused()
	     << F("\twS ") << stackUnused(taskStackBase, taskStackSize)
	     << F("\tmaxt: ") << maxt
	     << F("\r\nDelay: comp ") << lateReadMax[false]
	     << F("\tyield ") << lateReadMax[true] << endl;
      if (st == Debug)
	{
	  blinkSensor[0]->debug('p');
	  blinkSensor[0]->debug();
	  if (blinkSensor[1]) blinkSensor[1]->debug();
	  for (byte i = 0; i < numOther; i++) otherSensor[i]->debug();
	}
      st = reading;
      return;

    case reset:
      Serial << endl;
    case Reset:
      maxt = 0;
      resetDebugStats();		// reset debug statistics
      blinkSensor[0]->reset();
      if (blinkSensor[1]) blinkSensor[1]->reset();
      for (byte i = 0; i < numOther; i++) otherSensor[i]->reset();
#     if debugLed
        debugLedFlashLen = -debugLedPeriod;
#     endif
      st = reading;
      break;				// proceed with reading

    case reading:
      break;				// proceed with reading
    }

  // Wait for the right time to do the sampling, possibly yield to other tasks
  mainLoopScheduler(blinkSensor, otherSensor, numOther, csp, nsp);

  // Do the real work
  csp->readSensor();			// sample the sensor reading
  nsp->prepareRead();			// prepare for next sampling
  csp->publishReading();		// compute reading data and publish it
}

/*
 * Scheduling policy
 *
 * To obtain jitter-free readings, minTimeForComputingUs should be
 * greater than the maximum time needed by a reading, and minTimeForYieldingUs
 * should be greater than the maximum time other tasks will run without yielding.
 *
 * To optimise, use this algorithm:
 * minTimeForComputingUs = readingIntvUs / 3;
 * minTimeForYieldingUs = readingIntvUs / 3 - 150;
 * decrese minTimeForComputingUs until lateReadMax[false] > 20
 * decrease minTimeForYieldingUs until lateReadMax[true] > 20
 */
inline void mainLoopScheduler (AnalogSensor *blinkSensor[],
			       AnalogSensor *otherSensor[], const byte numOther,
			       AnalogSensor *&csp, AnalogSensor *&nsp)
{
  const int minTimeForComputingUs = 420;
  const int minTimeForYieldingUs = 720;

  // Current reading has been done: set interval
  // from current to next reading depending on type of current sensor
  const byte numBlink = (blinkSensor[1]) ? 2 : 1;
  static byte otherIndex = 0;		// index into other sensors array
  int toNextUs;				// distance from current to next reading
  csp = nsp;				// copy next sensor pointer to current one
  if (csp == blinkSensor[0]		// if current is first blink sensor
      && numBlink == 2)			// two blink sensors present
    {
      toNextUs = minTimeForComputingUs;
      nsp = blinkSensor[1];		// next is last blink sensor
    }
  else if (csp == blinkSensor[numBlink-1]) // if current is last blink sensor
    {
      toNextUs = minTimeForComputingUs;
      nsp = otherSensor[otherIndex];	// next is an other sensor
      if (++otherIndex == numOther)	// advance index of other sensor for next time
	otherIndex = 0;
    }
  else				// if current is an other sensor
    {
      toNextUs = readingIntvUs - numBlink * minTimeForComputingUs;
      nsp = blinkSensor[0];		// next is first blink sensor
    }

  // Decide whether to yield, depending on time left to next scheduled reading
  static int timeToReadUs = 0;
  int lateReadUs;			// DEBUG: statistics on scheduling delay
  bool yielding = false;		// note whether we yield
  timeToReadUs += toNextUs;		// this is the time for the next reading
  while ((timeToReadUs - (int)micros()) > minTimeForYieldingUs)
    {
      yielding = true;
      yield();				// let other tasks run
    }
  while ((lateReadUs = (int)micros() - timeToReadUs) < 0)
    ;					// wait without yielding
  lateReadMax[yielding] = max(lateReadMax[yielding], lateReadUs); // DEBUG
}


/********************* Web Server ************************/

bool serveWebConnection(Client&);
void answerClient(Stream&, char*);

bool persistentConnection = true;
//File webFile;				// the web page file on the SD card

WiFiServer server[] = { WiFiServer(81),
			WiFiServer(80) }; // web servers, one or two
const byte serverCount = sizeof(server) / sizeof(*server);
WiFiClient client[serverCount];

void webServerTask ()
{
  // When we connect to the WiFi network, init the server
  if (wifiStatus != WL_CONNECTED)	// we are disconnected
    {
      if ((wifiStatus = WiFi.begin((char *)ssid)) == WL_CONNECTED)
	{
	  Serial << F("W\r\n");
	  for (byte ci = serverCount; ci--; )
	    server[ci].begin();
	}
      else
	{
	  Serial << F("WiFi connect failed, reset\r\n"); delay(5);
	  resetWiFiShield();
	}
    }

  // Set the Unix time
  while (unixTimeAtStartup == 0)
    {
      WiFiClient tcp;
      time_t t = webUnixTime(tcp);
      delay(50);
      if (t)				// time was successfully retrieved
	{
	  unixTimeAtStartup = t * 1000ull - millis(); // note when we started
	  Serial + t + endl			      // print Unix time
	    + unixTimeAtStartup + endl + push; delay(5); // print startup time
	  resetDebugStats();		// reset debug statistics
	}
      else				// failed to get the time, let's try again
	{
	  Serial << F("F\r\n");
	  delay(2000);
	}
    }

  for (byte ci = serverCount; ci--; )
    if (client[ci])
      {					// a client is there waiting
	if (client[ci].connected()
	    && serveWebConnection(client[ci]))
	  {
	    delay(50);			   // everything went well
	    wifidogTimeout = millis() + wifidogAdvance; // timeout for stuck WiFi shield
	  }
	else
	  {
	    // Something went wrong: close the connection
	    Serial << F("C\r\n");
	    delay(50);
	    client[ci].stop();
	  }
      }
    else				// no client yet
      {
	client[ci] = server[ci].available();
	client[ci].setTimeout(5000);
	delay(50);
      }
}

bool serveWebConnection(Client& client)
{
  char command[8];
  byte nr;

  nr = client.readBytesUntil('/', command, sizeof(command));
  command[min(7,nr)] = '\0';

  if (nr > 0
      && strncmp_P(command, PSTR("GET "), 4) == 0)
    {
      nr = client.readBytesUntil('\r', command, sizeof(command));
      command[min(7,nr)] = '\0';

      if (nr > 0)
	{
	  // It is important to answer before reading the whole command, not after:
	  // it works both ways but this way it is much faster with the WiFi shield
	  // probably because of some timing internal to the shield.
	  answerClient(client, command); // answer before reading the whole request

	  char endOfHeaders[] = "\r\n\r\n";
	  if (client.find(endOfHeaders))
	    return true;
	}
    }

  return false;
}

void answerClient(Stream& str, char* command)
{
  str + clear + F("HTTP/1.1 200 OK\r\n"); // clear protects from programming errors
  if (! persistentConnection)
    str + F("Connection: close\r\n");
  str + F("Content-Type: text/");

  if (command[0] == ' ')
    {
      resetDebugStats();		// reset debug statistics on reload

      str +F("html\r\n"
// 	     "Content-Length: ");
//       webFile = SD.open("index.htm");
//       if (webFile)
// 	{
// 	  str + webFile.size() + endl + endl + push;
// 	  while(webFile.available())
// 	    {
// 	      str.write(b.buf, webFile.read(b.buf, sizeof(b.buf))); // send web page to client
// 	      yield();
// 	    }
// 	  webFile.close();
// 	}
//       else
// 	str + F("0\r\n\r\n") + push;
	      "Transfer-Encoding: chunked\r\n\r\n")
	+ chunk
	+ F(
"<!DOCTYPE html>\n"
"<HTML>\n"
  "<HEAD>\n"
    "<META charset='UTF-8'>\n"
    "<TITLE>Arduino leggiluce</TITLE>\n"
    "<SCRIPT>\n"
      "var recoverTimeout;\n"
      "function upd(xml,id,tag)\n"
      "{\n"
	  "var list = xml.getElementsByTagName(tag);\n"
	  "if (list.length > 0) {\n"
	      "var val = list[0].childNodes[0].nodeValue;\n"
	      "var element = document.getElementById(id);\n"
	      "if (element) {\n"
		  "if (id == 'date')\n"
		      "val = new Date(1000*parseInt(val));\n"
		  "element.innerHTML = val;\n"
	      "}\n"
	  "}\n"
      "}\n"
      "function readW()\n"
      "{\n"
	  "var request = new XMLHttpRequest();\n"
	  "request.onreadystatechange = function() {\n"
	      "if (this.readyState == 4\n"
		  "&& this.status == 200\n"
		  "&& this.responseXML != null) {\n"
		  "var xml = this.responseXML;\n"
		  "upd(xml,'date','T');\n"
		  "upd(xml,'pin0','P0');\n"
		  "upd(xml,'pin1','P1');\n"
		  "upd(xml,'uptime','UT');\n"
		  "upd(xml,'stackunused','WS');\n"
		  "upd(xml,'delayc','DC');\n"
		  "upd(xml,'delayy','DY');\n"
		  "upd(xml,'readingslost','RL');\n"
		  "clearTimeout(recoverTimeout);\n"
		  "setTimeout('readW()', 600);\n"
	      "}\n"
	  "}\n"
	  "request.open('GET', 'ajaxReq', true);\n"
	  "request.send(null);\n"
	  "recoverTimeout = setTimeout('readW()', 150000);\n"
      "}\n"
    "</SCRIPT>\n"
  "</HEAD>\n"
  "<BODY onload='readW()'>\n"
    "<H1>Lettura ENEL</H1>\n"
    "<P>Potenza istantanea: <SPAN id=pin0>...</SPAN> watt</P>\n"
    "<SMALL>\n"
    "<P>Potenza reattiva: <SPAN id=pin1>...</SPAN> VA</P>\n"
    "<TABLE border=1>\n"
      "<TR><TH colspan=3>Debugging data\n"
      "<TR><TD id=date align=center colspan=3>\n"
      "<TR><TD>uptime<TD id=uptime align=right>...<TD>seconds</TR>\n"
      "<TR><TD>webServer task unused stack<TD id=stackunused align=right>...<TD>bytes</TR>\n"
      "<TR><TD>late read because of computation<TD id=delayc align=right>...<TD>µs</TR>\n"
      "<TR><TD>late read because of yield<TD id=delayy align=right>...<TD>µs</TR>\n"
      "<TR><TD>readings lost<TD id=readingslost align=right>...<TD>times</TR>\n"
    "</TABLE>\n"
    "</SMALL>\n"
  "</BODY>\n"
"</HTML>\n"
	    )
	+ push;
    }

  else if (strncmp_P(command, PSTR("ajaxReq"), 7) == 0)
    {
      reading r;

      // Ajax request - send XML file
      str + F("xml\r\n"
	      // If cross-script does not work: http://www.html5rocks.com/en/tutorials/cors/
	      // Keep-Alive: timeout=2, max=100
	      // Content-Type: application/xml
	      "Access-Control-Allow-Origin: *\r\n"
	      // Access-Control-Allow-Methods: POST, GET, PUT, DELETE
	      // Access-Control-Allow-Headers: Authorization
	      "Cache-Control: no-cache\r\n"
	      "Transfer-Encoding: chunked\r\n\r\n")
	+ chunk
	+ F("<?xml version='1.0'?><_>");

      if (disaster)			// signal disaster, do nothing else
	{
	  str + F("<DISASTER></DISASTER>");
	}
      else
	{
	  while (readings.get(r))
	    {
	      if (readingLost)
		{
		  str + F("<RL>") + int(readingLost) + F("</RL>");
		  readingLost = 0;
		}

	      switch (r.type)
		{
		case '0':
		case '1':
		  {
		    char idTail[] = { r.type, '>', '\0' };
		    str + F("<P") + idTail + r.val + F("</P") + idTail
		      + F("<T") + idTail + r.time + F("</T") + idTail;
		  }
		  break;
		case 'S':
		  str + F("<RS></RS>");
		  break;
		}
	    }

	  str + F("<UT>") + millis()/1000 + F("</UT>"
	    "<WS>") + int(stackUnused(taskStackBase, taskStackSize)) + F("</WS>"
	    "<DC>") + lateReadMax[false] + F("</DC>"
	    "<DY>") + lateReadMax[true] + F("</DY>");
	}

      str + F("</_>\n") + push;
    }
}


/************** Sensor class implementation ***************/

AnalogSensor::AnalogSensor (byte p) : pin(p)
{
  pinMode(A0+pin, INPUT);
  digitalWrite(A0+pin, HIGH);

  reset();
}

BlinkSensor::BlinkSensor (byte p) : AnalogSensor (p)
{
  initProto();
}

BlinkSensor::~BlinkSensor ()
{
  free(proto);
}

void BlinkSensor::reset ()
{
  valRSum = valRSumSq = 0;
  corIdx = 0;
  memset(corWind, 0, sizeof(corWind));
  blinkCor = 0;
  blinkVarEnvelope = 0;
  prevBlinkTimeMs = 0;
}

// Initialize prototypes
void BlinkSensor::initProto ()
{
  // This is the definition of the blink prototypes used by the correlator.
  // The prototypes are corLen long, and have a low value for all but a
  // central spike of length blinkWidth and height highMinusLowBlink+lowBlink.
  // The average of a prototype (blink plus tails) is 0.
  // Prototypes are stored only for the blink part, that is, starting from
  // when the blink starts and extending up to corLen/2.

  // proto is the structures holding the blink part of prototypes increased by
  // lowblink, so it contains only positive values which should be added to
  // lowBlink, which is negative, when making up a whole prototype.
  // proto stores the blink part of prototypes sequentially, each terminated by 0.

  // Check if prototype has already been initialised
  if (proto)
    return;

  // Initialise the prototype blink (central part) as a smeared rectangular pulse
  // The central rectangular spike is smeared using a one-pole filter with
  // time constant alpha.  Three values of alpha are considered, corresponding to
  // time constants of 5, 12 and 30 ms, matching the delayed response typical of
  // many photoresistors.  alpha set to 1 disables smearing.
  // Time constants should be in increasing order, for memory and time efficiency.
  // assert(protoNum == sizeof(timeConst)/sizeof(*timeConst);
  const byte timeConst[] = { 5, 12, 30 }; // time constants [ms]
  int allocated = 0;				      // proto size
  for (byte pidx = 0; pidx < protoNum; pidx++)
    {
      // Allocate storage for proto, a list of zero-terminated blink prototypes
      // We rely on realloc to extend the previously allocated space and we do not
      // check for a valid returned pointer.
      const byte maxBlinkProtoLen = corLen/2 + blinkWidth/2;
      proto = (byte *)realloc(proto, allocated + maxBlinkProtoLen + 1);

      // Find half the highest shift that does not overflow corLen << scale
      //const byte scale = (8*sizeof(long) - 2 - ceil(log(corLen)/log(2))) / 2;
      const byte scale = 8;		// this is easier, faster, smaller, slightly less accurate
      long alpha = (1000l << scale) / readingsPerPowerPeriod / powerFrequency / timeConst[pidx];
      long magic = 0;			// force a prototype with null sum
      byte inc = 2*scale;		// bisection increment
      while (true)
	{
	  long y = 0;
	  int sum = lowBlink * corLen;
	  for (byte i = 0; i < maxBlinkProtoLen; i++)
	    {
	      const long rounding = 1l << (2*scale-1);
	      long x = (i < blinkWidth) ? (long)highMinusLowBlink << scale : 0;
	      y += (x-(y >> scale)) * alpha;
	      sum += proto[allocated+i] = (y + magic + rounding) >> 2*scale;
	    }
	  if (sum == 0)
	    break;				    // magic number found
	  magic -= ((sum > 0) ? 1l : -1l) << --inc; // adjust magic and recompute proto
	}

      // Compute blink prototype length, sum of squares and store multiplier
      byte protoLen =  maxBlinkProtoLen;       // start from the end of proto
      while (proto[allocated+protoLen-1] == 0) // look behind to skip trailing zeroes
	protoLen -= 1;			       // keep going back until we find the tail
      unsigned protoSumSq = lowBlink * lowBlink * (corLen - protoLen);
      for (byte i = 0; i < protoLen; i++)
	{
	  byte protoVal = lowBlink + proto[allocated+i];
	  protoSumSq += protoVal * protoVal;
	}
      proto[allocated + protoLen] = 0;	// add terminator
      protoMult[pidx] = 256*256ul * corLen / protoSumSq;

      // Free unused memory after tail
      // The +2 is there to work around a bug in realloc() shrinking memory,
      // which corrupts the two last bytes of the returned shrunk chunk.
      // The bug was fixed in avr-libc 1.7.0, and in fact the Arduino environment
      // 1.05 to 1.5.4 at least includes realloc.c from avr-libc 1.8.0, but if for any
      // reason you happen to use avr-libc prior to 1.7.0 this will save your day...
      const byte avr_libc_workaround = 2; // you can zero this out if avr-libc >= 1.7.0
      proto = (byte *)realloc(proto, (allocated += protoLen + 1) + avr_libc_workaround);
    }
}

void AnalogSensor::prepareRead ()
{
  /*
   * One way is to make one reading and discard the results to minimise
   * interference from previous analog readings on other inputs; this works
   * because the first reading sets the ADMUX register, thus connecting the
   * sensor to the ADC input capacitance and giving it the time to charge it.
   * The most efficient way is to just set the ADMUX register, do something else
   * and after some time come back and sample.
   */

  // Set ADMUX, do not forget to set the analog reference
  ADMUX = (DEFAULT << 6) | pin;
  ADMUX = (INTERNAL << 6) | 8;
}

void BlinkSensor::prepareRead ()
{
  // Set ADMUX, do not forget to set the default (Vcc) analog reference
  ADMUX = (DEFAULT << 6) | pin;
}

void AnalogSensor::readSensor (byte prescaler)
{
  // The ADC clock is the chip clock (16 MHz on Arduino Uno) divided by
  // the prescaler, whose value is 2^prescaler, so for the default prescaler
  // value of 7 the ADC clock is 16/128 MHz.  The conversion lasts 13 cycles.
  ADCSRA &= ~0x07;			// Clear the three prescaler bits
  ADCSRA |= prescaler & 0x07;		// Set the prescaler (only three low bits)

  // Start the conversion
  bitSet(ADCSRA, ADSC);

  // ADSC is cleared when the conversion finishes
  while (bitRead(ADCSRA, ADSC));

  // Read ADCL first.
  val = ADCL | ADCH << 8;
}

void BlinkSensor::readSensor (byte prescaler)
{
  // No need for high accuracy: default is prescaler 8 so ADC clock speed 2 MHz

  // When called through a pointer to the base calss AnalogSensor, the default
  // value of the base class is used for prescaler (which is 8), rather than the
  // one of the derived BlinkSensor class (which is 3).  So when calling the base
  // class function we use a literal instead of the function argument.
  // AnalogSensor::readSensor(prescaler);
  AnalogSensor::readSensor(3);

  // Lower resolution and meaning
  readVal = 256 - (val >> 2);
}

// Compute covariance of readings with blink prototype number i
int BlinkSensor::covariance (byte pidx)
{
  register int cov = valRSum;
  cov *= lowBlink;
  const int blinkCenter = corLen/2;
  byte corProtoStartIdx = corIdx + blinkCenter - blinkWidth/2;
  register byte *cp = corWind + corProtoStartIdx;
  register byte *pp = proto;		// pointer to walk proto
  for (byte i = 0; i < pidx; i++)	// walk the previous blink prototypes
    while (*pp++) ;			// walk until terminator 0
  while (byte protoval = *pp++)
    {
      if (cp >= corWind+corLen)
	cp -= corLen;
      cov += (int)*cp++ * protoval;
    }
  return cov;
}

void BlinkSensor::publishReading ()
{
  const long lowDev = 5;		// minimum deviation expected for a blink
  const unsigned long lowVar = lowDev*lowDev * blinkWidth * corLen;
  unsigned long valRVar = corLen * valRSumSq - (unsigned long)valRSum * valRSum;

  // Update running sum, running sum of squared values
  // Compute running variance
  valRSum += readVal;
  valRSum -= corWind[corIdx];
  valRSumSq += (unsigned)readVal*readVal;
  valRSumSq -= (unsigned)corWind[corIdx]*corWind[corIdx];

  // Update history and increment index with wraparound
  corWind[corIdx] = readVal;
  if (++corIdx >= corLen)
    corIdx = 0;


  /***************** Compute correlations *****************/
  if (valRVar > lowVar)			// if variance too low let cor be 0
    // Compute correlations with protoNum different blink signal prototypes.
    // If positive, store the highest squared correlation in cor
    {
      unsigned sqCor = 0;			// maximum of the squared correlations

      // TIMING: the following loop needs 224 µs
      for (byte pidx = 0; pidx < protoNum; pidx++)
	{
	  // Compute covariance with prototype
	  // TIMING: this call needs 70 µs in the worst case
	  register int cov = covariance(pidx);

	  // Compute squared correlation, if covariance is positive
	  // Also compute variance, for envelope detector
	  // TIMING: the following branch taken needs 64 µs (40 with floating point math)
	  if (cov > 0)
	    {
	      // correlation is square(cov) * corLen / (valRVar * protoSumSq[pidx])
	      // protoMult is 256 * 256 * corLen / protoSumSq
	      // thisCor256 is 256 * correlation
	      unsigned long thisCor256 = protoMult[pidx] * cov >> (8 + 2);
	      thisCor256 *= cov;
	      thisCor256 /= valRVar >> 2;

	      // DEBUG: check that integer arithmetic is the same as floating point
	      if (false)
		{
		  unsigned long thisCor256d = square(cov) * protoMult[pidx] / valRVar / 256;
		  if (thisCor256d != thisCor256)
		    Serial << protoMult[pidx] << '\t' << cov << '\t' << valRVar << endl
			   << '\t' << thisCor256 << '\t' << thisCor256d << endl;
		}

	      // DEBUG: if squared correlation > 1 something terrible is happening!
	      if (thisCor256 > 255)
		{
		  disaster = true;
		  Serial << endl
			 << F("ERROR: sqCor=") << thisCor256 << F(" on proto ") << pidx << endl
			 << F("cov=") << cov << F(" valRVar=") << valRVar << endl;
		}

	      sqCor = max(sqCor, unsigned(thisCor256));
	    }
	}

      // Only compute variance and correlation if something significant seen
      if (sqCor < 128)			// squared correlation < 0.5
	cor = 0;			// nothing seen
      else				// squared correlation >= 0.5
	{
	  // Scale the variance with respect to the minimum expected variance for a blink,
	  // approximated to a power of 2, then inflate it as much as possible without
	  // overflowing var so that the envelope detector has enough room to do its job.
	  const byte lowVarLog2 = ceil(log(lowVar)/log(2));
	  var = min(maxuint, valRVar >> (lowVarLog2 - 5));
	  cor = sqCor;
	}
    }


  /********************* Identify blink *******************/
  const byte highThreshold = 255 - 50;		 // correlation threshold
  const byte lowThreshold = highThreshold - 25;	 // hysteresis

  // As soon as the correlation exceeds the high threshold, start following it
  // until the maximum value is reached
  if (cor > highThreshold)
    {
      if (cor > blinkCor)
	{
	  blinkCor = cor;		// blink starts
	  blinkVar = var;
	  blinkTimeMs = millis();
	}
    }

  // Once the high threshold has been exceeded, wait for the correlation to
  // fall below the low threshold, then signal that a blink has been detected
  else if (blinkCor > 0 && cor < lowThreshold)
    {
      // TIMING: this if branch needs at most 100 µs
      // Variance envelope detector
      const byte varEnvFraction = 4;		    // best if a power of two
      const byte varEnvDecay = 128;		    // best if a power of two
      blinkVarEnvelope -= blinkVarEnvelope / varEnvDecay;
      if (blinkVar > blinkVarEnvelope)	// follow envelope
	blinkVarEnvelope = blinkVar;

      // If power is high enough, blink is detected
      if (blinkVar > blinkVarEnvelope / varEnvFraction)
	{
	  if (prevBlinkTimeMs != 0	 // do not signal first blink
	      && unixTimeAtStartup != 0) // signal blink only if time was set
	    {
	      // Set measured power
	      millis_t interval = blinkTimeMs-prevBlinkTimeMs;
	      int power = (1 + 2 * 3600 * 1000l / interval) / 2; // rounding

	      // Set time of blink
	      unsigned long long time = unixTimeAtStartup;
	      const bool stepGraphHelper = true; // use previous time
	      if (stepGraphHelper)	// facilitate drawing step graphs
		time += prevBlinkTimeMs;
	      else
		time += blinkTimeMs;

	      // Put blink in FIFO
	      reading blink = { '0'+pin, time, power };
	      if (! readings.put(blink)) // add new data and check if FIFO full
		{
		  reading r;		// trash
		  readings.get(r);	// discard the oldest reading
		  readingLost += 1;	// note one more discarded item
		  readings.put(blink);	// put in the most recent one
		}
	    }

	  prevBlinkTimeMs = blinkTimeMs;
	}

      blinkCor = 0;			// blink ends
    }
}

void AnalogSensor::debug (char command, int arg)
{
  Serial << F("--> ")  << pin << endl
	 << F("val: ") << val << endl;
}

void InternalTempSensor::debug (char command, int arg)
{
  // From data sheet, slope is 1.062 mv/°C, 0°C is typically at 287 mV.
  // Using the internal 1.1 V reference, we read 1.075 mV/bit,
  // so 0°C reads as 267 bit and slope is 1.012 °C/bit (259/256)
  // Real offset and real slope are very different
  const int offset = 92;
  const int gainNum = 210;
  const int gainDen = 256;		// 256/210=1.22
  int temp = (val-offset) * gainNum / gainDen;
  Serial << F("--> ") << pin
	 << F("\tTemp: ") << temp << F("°C\r\n");
}

void BlinkSensor::debug (char command, int arg)
{
  switch (command) {

  case 'p':
    for (byte pidx = 0; pidx < protoNum; pidx++) {
      int sumproto = - corLen;
      Serial << F("proto") << pidx << F("=[");
      byte *pp = proto;			  // pointer to walk proto
      for (byte i = 0; i < pidx; i++)	  // walk the previous blink prototypes
	while (*pp++) ;			  // walk this blink prototype until terminator 0
      byte *protoStart = pp;		  // use this to compute length of blink prototype
      while (byte protoVal = *pp++)
	{
	  Serial << ' ' << protoVal;
	  sumproto += protoVal;
	}
      Serial << F("]; len=") << pp - protoStart - 1
	     << F(" sum=") << sumproto << F(" protoMult=") << protoMult[pidx] << endl;
    }
  break;
  case 'n':
    AnalogSensor::debug(command, arg);
    break;
  }
}


/****************** Utility functions *********************/

// call this without arguments to light a led, or with argument 0 to turn it down
// Call this from within loop() to blink it with n. of ms to switch it on and off
void ledOnOff (int pin, unsigned on, unsigned off)
{
  int newstate;
  if (on == LOW)
    newstate = LOW;
  else if (off == LOW)
    newstate = HIGH;
  else if ((unsigned)millis() % (on+off) < on)
    newstate = HIGH;
  else
    newstate = LOW;
  digitalWrite(pin, newstate);
}

time_t webUnixTime (Client &client)
{
  time_t time = 0;

  // Just choose any reasonably busy web server, the load is really low
  if (client.connect("g.cn", 80))
    {
      yield();

      // Make an HTTP 1.1 request which misses a Host: header
      // Compliant servers are required to answer with an error that includes
      // a Date: header.
      client + F("GET / HTTP/1.1\r\n\r\n") + push;

      char buf[5];			    // temporary buffer for characters
      client.setTimeout(15000);		    // timeout for reading answer [ms]
      if (client.find((char *)"\r\nDate: ") // look for Date: header
	  && client.readBytes(buf, 5) == 5) // discard
	{
	  yield();

	  unsigned day = client.parseInt(); // day
	  client.readBytes(buf, 1);	   // discard
	  client.readBytes(buf, 3);	   // month
	  int year = client.parseInt();	   // year
	  byte hour = client.parseInt();   // hour
	  byte minute = client.parseInt(); // minute
	  byte second = client.parseInt(); // second

	  int daysInPrevMonths;
	  switch (buf[0])
	    {
	    case 'F': daysInPrevMonths =  31; break; // Feb
	    case 'S': daysInPrevMonths = 243; break; // Sep
	    case 'O': daysInPrevMonths = 273; break; // Oct
	    case 'N': daysInPrevMonths = 304; break; // Nov
	    case 'D': daysInPrevMonths = 334; break; // Dec
	    default:
	      if (buf[0] == 'J' && buf[1] == 'a')
		daysInPrevMonths = 0;		// Jan
	      else if (buf[0] == 'A' && buf[1] == 'p')
		daysInPrevMonths = 90;		// Apr
	      else switch (buf[2])
		     {
		     case 'r': daysInPrevMonths =  59; break; // Mar
		     case 'y': daysInPrevMonths = 120; break; // May
		     case 'n': daysInPrevMonths = 151; break; // Jun
		     case 'l': daysInPrevMonths = 181; break; // Jul
		     default: // add a default label here to avoid compiler warning
		     case 'g': daysInPrevMonths = 212; break; // Aug
		     }
	    }

	  // This code will not work after February 2100
	  // because it does not account for 2100 not being a leap year and because
	  // we use the day variable as accumulator, which would overflow in 2149
	  day += (year - 1970) * 365;	// days from 1970 to the whole past year
	  day += (year - 1969) >> 2;	// plus one day per leap year
	  day += daysInPrevMonths;	// plus days for previous months this year
	  if (daysInPrevMonths >= 59	// if we are past February
	      && ((year & 3) == 0))	// and this is a leap year
	    day += 1;			// add one day
	  // Remove today, add hours, minutes and seconds this month
	  time = (((day-1ul) * 24 + hour) * 60 + minute) * 60 + second;
	}
    }
  delay(10);
  client.flush();
  client.stop();
  yield();

  return time;
}

// Use this rather than Scheduler.startLoop() and save the pointer returned
// so that you can  pass it to stackUnused().
// This function relies on malloc to just give back the same pointer after free(),
// and on Scheduler.StarLoop() to do the first malloc() call to allocate the stack.
// If the above is not true, memory corruption will result.
byte *startLoop (SchedulerTask task, unsigned stackSize) {
  byte *stack = (byte *)malloc(stackSize); // the same allocation Scheduler.StarLoop will do
  // If we are being called from the main loop, let's fill our stack, too
  extern byte *__brkval;		   // set after malloc, it is the top of the heap
  if ((byte *)&stack > __brkval)	   // this function is called from the main loop
    for (byte *end = (byte *)&end - 10;	   // leave space for a small interrupt routine
	 __brkval < end;		   // until end pointer smashes into the heap
	 *end-- = 0x55) ;		   // fill stack with fixed value
  memset(stack, 0x55, stackSize);	   // fill the stack's stack with fixed values
  free(stack);				   // now we free the space allocate with malloc
  Scheduler.startLoop(task, stackSize);	   // start the task
  memset(stack, 0x55, 4);		   // rewrite over memory dirtied by free()
  return stack;				   // stack base, to be passed to stackUnused()
}

// When not using the scheduler, call it from setup()
// Else, this work is done when calling startLoop()
void fillStack () {
  extern byte __heap_start, *__brkval;
  byte *beg = (byte*)(__brkval ? __brkval : &__heap_start);
  byte *end = (byte*)&end - 10;		// leave space for a small interrupt routine
  while (beg < end) *beg++ = 0x55;	// can't use memset here: it would use the stack
}

// Count bytes marked 0x55 from the stack base to the the local variables
// Should be called with either no arguments, for counting free stack on the main loop,
// or with two arguments, for the tasks.
unsigned stackUnused (byte *stackBase, unsigned stackSize) {
  byte *end = stackSize ? stackBase + stackSize : (byte *)&end - 10;
  byte *p = stackBase;
  while (*p == 0x55 && p < end)
    p += 1;
  return p - stackBase;
}

void resetDebugStats ()
{
  const reading NaNblink = { 'S' };

  memset(lateReadMax, 0, sizeof(lateReadMax));
  readings.flush();			// clear FIFO readings entries
  readings.put(NaNblink);		// mark separation
}


/*
 * This code assumes that at least one of the following is true:
 * 1) R7 on WiFi shield is shorted (see schematics)
 * 2) both the RESET pin of the WiFi shield and PIN 5 on the ICSP connector on
 * the board below the WiFi shield are bent so they are not connected and in
 * addition a wire on the WiFi shield connects A5 and RESET; in this case,
 * manual reset buttons are disconnected; moreover, a call to resetWiFiShield()
 * in setup() is necessary to reset the shield at startup.
 */
void resetWiFiShield ()
{
  pinMode(resetPin, OUTPUT);		// set output: low is the default state
  delayMicroseconds(10);		// ADM811 input pulse should be 10 µs long
  pinMode(resetPin, INPUT);		// set to input mode
}

void resetArduino ()
{
  while (usingWatchdogTimer)		// is the wat dog unchained?
    ;					// just wait for it to byte
					// else
  void (*restart) () = 0;		// function pointer to address 0
  restart();				// jump there to restart the system
}

// Local Variables:
// fill-column: 100
// c-macro-cppflags: "-include Arduino.h -MMD -mmcu=atmega328p -DF_CPU=16000000L -DARDUINO=154 -I. -I../arduino-1.5.4/hardware/arduino/cores/arduino -I../arduino-1.5.4/hardware/arduino/variants/standard -I../arduino-1.5.4/libraries/SPI -I../arduino-1.5.4/libraries/WiFi -I../arduino-1.5.4/libraries/SD -I../arduino-1.5.4/libraries/SPI -I../arduino-1.5.4/libraries/WiFi/utility -I../arduino-1.5.4/libraries/SD/utility -I/home/pot/src/arduino/libraries/Streaming -I/home/pot/src/arduino/libraries/SchedulerARMAVR"
// End:
